/*
  */

#include "ocr.h"

typedef struct {
  Pixcol col;
  struct DatabaseNode *then;
} DatabaseLink;

#define MAXGLYPHCHRS 3

typedef struct DatabaseNode {
  char s[MAXGLYPHCHRS+1]; /* null-terminated; "" means no match here */
  int nlinks, alinks;
  DatabaseLink *links;
} DatabaseNode;

static const char *context_names[]= {
  "Lower",
  "Upper",
  "Digit"
};

#define NCONTEXTS (sizeof(context_names)/sizeof(context_names[0]))

struct OcrReader {
  int h;
  DatabaseNode contexts[NCONTEXTS];
  OcrResultGlyph *results;
  int aresults, nresults;
};

static FILE *resolver;
static pid_t resolver_pid;
static int resolver_done;

static void fgetsline(FILE *f, char *lbuf, size_t lbufsz) {
  char *s= fgets(lbuf,lbufsz,f);
  eassert(s);
  int l= strlen(lbuf);
  eassert(l>0);  eassert(lbuf[--l]='\n');
  lbuf[l]= 0;
}
#define FGETSLINE(f,buf) (fgetsline(f,buf,sizeof(buf)))

static void cleardb_node(DatabaseNode *n) {
  int i;
  n->s[0]= 0;
  for (i=0; i<n->nlinks; i++)
    cleardb_node(n->links[i].then);
}

static void readdb(OcrReader *rd) {
  int nchrs;
  DatabaseNode *current, *additional;
  char chrs[MAXGLYPHCHRS+1];
  Pixcol cv;
  int r,j,ctxi;
  int h;
  char lbuf[100];
  FILE *db;

  char *dbfname=0;
  asprintf(&dbfname,"%s/charset-%d.txt",get_vardir(),rd->h);
  eassert(dbfname);
  
  db= fopen(dbfname,"r");  eassert(db);
  free(dbfname);

  FGETSLINE(db,lbuf);
  eassert(!strcmp(lbuf,"# ypp-sc-tools pctb font v1"));

  r= fscanf(db, "%d", &h);
  eassert(r==1);
  eassert(h==rd->h);

  for (ctxi=0; ctxi<NCONTEXTS; ctxi++)
    cleardb_node(&rd->contexts[ctxi]);

  for (;;) {
    FGETSLINE(db,lbuf);
    if (!lbuf || lbuf[0]=='#') continue;
    if (!strcmp(lbuf,".")) break;

    for (ctxi=0; ctxi<NCONTEXTS; ctxi++)
      if (!strcmp(lbuf,context_names[ctxi]))
	goto found_ctx;
    /* not found, just skip */
    for (;;) { FGETSLINE(db,lbuf); if (!lbuf[0]) break; }
    continue;

  found_ctx:
    for (nchrs=0;;) {
      int c= fgetc(db);  eassert(c!=EOF);
      if (c=='\n') { eassert(nchrs); break; }
      eassert(nchrs<MAXGLYPHCHRS);
      if (c=='\\') {
	unsigned cr;
	c= fgetc(db);  eassert(c=='x');
	r= fscanf(db, "%2x", &cr);  eassert(r==1);
	assert(cr>0 && cr<=255);
	c= cr;
      }
      chrs[nchrs++]= c;
    }
    chrs[nchrs]= 0;

    current= &rd->contexts[ctxi];
    for (;;) {
      FGETSLINE(db,lbuf);
      if (!lbuf[0]) { eassert(current != &rd->contexts[ctxi]); break; }
      char *ep;
      cv= strtoul(lbuf,&ep,16);  eassert(!*ep);
      eassert(!(cv & ~((1UL << rd->h)-1)));
      
      for (j=0; j<current->nlinks; j++)
	if (current->links[j].col == cv) {
	  current= current->links[j].then;
	  goto found_link;
	}

      additional= malloc(sizeof(*additional)); eassert(additional);
      additional->s[0]= 0;
      additional->nlinks= additional->alinks= 0;
      additional->links= 0;
      if (current->nlinks==current->alinks) {
	current->alinks++;
	current->alinks<<=1;
	current->links= realloc(current->links,
				sizeof(*current->links) * current->alinks);
	eassert(current->links);
      }
      current->links[current->nlinks].col= cv;
      current->links[current->nlinks].then= additional;
      current->nlinks++;
      current= additional;

    found_link:;
    }

    eassert(!current->s[0]);
    strcpy(current->s, chrs);
  }
  eassert(!ferror(db));
  eassert(!fclose(db));
}

static void cu_pr_ctxmap(unsigned ctxmap) {
  fprintf(resolver,"{");
  const char *spc="";
  int ctxi;
  for (ctxi=0; ctxi<NCONTEXTS; ctxi++) {
    if (!(ctxmap & (1u << ctxi))) continue;
    fprintf(resolver,"%s%s",spc,context_names[ctxi]);
    spc=" ";
  }
  fprintf(resolver,"}");
}

static void callout_unknown(OcrReader *rd, int w, Pixcol cols[],
			    int unk_l, int unk_r, unsigned unk_ctxmap) {
  int jobpipe[2],donepipe[2], c, r,i, x,y;
  const OcrResultGlyph *s;
  const char *p;
  char cb;
  Pixcol pv;
  
  if (!resolver) {
    r= pipe(jobpipe);  eassert(!r);
    r= pipe(donepipe);  eassert(!r);
    resolver_pid= fork();
    eassert(resolver_pid!=-1);
    if (!resolver_pid) {
      r= dup2(jobpipe[0],0); eassert(r==0);
      r= close(jobpipe[1]); eassert(!r);
      r= close(donepipe[0]); eassert(!r);
      /* we know donepipe[1] is >= 4 and we have dealt with all the others
       * so we aren't in any danger of overwriting some other fd 4: */
      r= dup2(donepipe[1],4); eassert(r==4);
      execlp("./show-thing.tcl", "./show-thing.tcl",
	     "--automatic","1",(char*)0);
      eassert(!"execlp failed");
    }
    r= close(jobpipe[0]); eassert(!r);
    r= close(donepipe[1]); eassert(!r);
    resolver= fdopen(jobpipe[1],"w"); eassert(resolver);
    resolver_done= donepipe[0];
  }
  fprintf(resolver,"%d %d ",unk_l,unk_r);
  cu_pr_ctxmap(unk_ctxmap);
  for (i=0, s=rd->results; i<rd->nresults; i++, s++) {
    if (!strcmp(s->s," ")) continue;
    fprintf(resolver," %d %d ",s->l,s->r);
    cu_pr_ctxmap(s->ctxmap);
    fprintf(resolver," ");
    for (p=s->s; (c= *p); p++) {
      if (c=='\\') fprintf(resolver,"\\%c",c);
      else if (c>=33 && c<=126) fputc(c,resolver);
      else fprintf(resolver,"\\x%02x",(unsigned char)c);
    }
  }
  fputc('\n',resolver);

  fprintf(resolver,
	  "/* XPM */\n"
	  "static char *t[] = {\n"
	  "/* columns rows colors chars-per-pixel */\n"
	  "\"%d %d 2 1\",\n"
	  "\"  c black\",\n"
	  "\"o c white\",\n",
	  w,rd->h);
  for (y=0, pv=1; y<rd->h; y++, pv<<=1) {
    fputc('"',resolver);
    for (x=0; x<w; x++)
      fputc(cols[x] & pv ? 'o' : ' ', resolver);
    fputs("\",\n",resolver);
  }
  fputs("};\n",resolver);
  eassert(!ferror(resolver));
  eassert(!fflush(resolver));

  eassert(resolver);

  for (;;) {
    r= read(resolver_done,&cb,1);
    if (r==-1) { eassert(errno==EINTR); continue; }
    break;
  }

  if (r==0) {
    pid_t pid;
    for (;;) {
      pid= waitpid(resolver_pid, &r, 0);
      if (pid==-1) { eassert(errno==EINTR); continue; }
      break;
    }
    eassert(pid==resolver_pid);
    if (WIFEXITED(r)) {
      eassert(!WEXITSTATUS(r));
      fclose(resolver);
      close(resolver_done);
      resolver= 0;
    } else if (WIFSIGNALED(r)) {
      eassert(!"resolver child died due to signal");
    } else {
      eassert(!"weird wait status");
    }
  } else {
    eassert(r==1);
    eassert(cb==0);
  }

  readdb(rd);
}

static void add_result(OcrReader *rd, const char *s, int l, int r,
		       unsigned ctxmap) {
  if (rd->nresults >= rd->aresults) {
    rd->aresults++; rd->aresults<<=1;
    rd->results= realloc(rd->results,sizeof(*rd->results)*rd->aresults);
    eassert(rd->results);
  }
  rd->results[rd->nresults].s= s;
  rd->results[rd->nresults].l= l;
  rd->results[rd->nresults].r= r;
  rd->results[rd->nresults].ctxmap= ctxmap;
  rd->nresults++;
}

struct OcrCellTypeInfo {
  unsigned initial, nextword, midword;
};
const struct OcrCellTypeInfo ocr_celltype_number= {
  4,4,4
};
const struct OcrCellTypeInfo ocr_celltype_text= {
  .initial=2 /* Uppercase */,
  .nextword=3 /* Either */,
  .midword=1 /* Lower only */
};

OcrResultGlyph *ocr(OcrReader *rd, OcrCellType ct, int w, Pixcol cols[]) {
  int nspaces;
  unsigned ctxmap;
  int ctxi, i, x;

 restart:

  nspaces=- w;
  ctxmap= ct->initial;
  rd->nresults=0;
  fprintf(debug,"OCR h=%d w=%d",rd->h,w);
  for (x=0; x<w; x++) fprintf(debug," %"PSPIXCOL(PRIx),cols[x]);
  fprintf(debug,"\n");
  debug_flush();

  x=0;
  for (;;) {
    debug_flush();
    /* skip spaces */
    if (x>=w)
      break;

    if (!cols[x]) {
      nspaces++;
      x++;
      if (nspaces==3) {
	fprintf(debug,"OCR  x=%x nspaces=%d space\n",x,nspaces);
	add_result(rd," ",x-nspaces,x+1,0);
	ctxmap= ct->nextword;
      }
      continue;
    }
    nspaces=0;

    /* find character */
    int lx=x;

    DatabaseNode *uniquematch= 0;
    int uniquematch_rx=-1;
    
    fprintf(debug,"OCR  lx=%d ctxmap=%x  ",lx,ctxmap);

    for (ctxi=0; ctxi<NCONTEXTS; ctxi++) {
      DatabaseNode *current= &rd->contexts[ctxi];;
      DatabaseNode *bestmatch= 0;
      int bestmatch_rx=-1;

      x= lx;
      if (!(ctxmap & (1u << ctxi))) continue;
      fprintf(debug," || %s",context_names[ctxi]);

      for (;;) {
	debug_flush();
	fprintf(debug," | x=%d",x);
	if (x>w) break;
	Pixcol cv= cols[x];
	fprintf(debug," cv=%"PSPIXCOL(PRIx),cv);
	for (i=0; i<current->nlinks; i++)
	  if (current->links[i].col == cv)
	    goto found;
	/* not found */
	fprintf(debug," ?");
	break;

      found:
	current= current->links[i].then;
	if (current->s[0]) {
	  fprintf(debug," \"%s\"",current->s);
	  bestmatch= current;
	  bestmatch_rx= x;
	} else {
	  fprintf(debug," ...");
	}

	x++;
      }
      
      if (bestmatch) {
	if (uniquematch && strcmp(bestmatch->s, uniquematch->s)) {
	  fprintf(debug, " ambiguous");
	  uniquematch= 0;
	  break;
	}
	uniquematch= bestmatch;
	uniquematch_rx= bestmatch_rx;
      }
    }

    if (uniquematch) {
      fprintf(debug," || YES\n");
      add_result(rd, uniquematch->s, lx, uniquematch_rx, ctxmap);
      x= uniquematch_rx+1;
      ctxmap= ct->midword;
    } else {
      int rx;
      fprintf(debug," || UNKNOWN");
      for (rx=lx; rx<w && cols[rx]; rx++);
      fprintf(debug," x=%d ctxmap=%x %d..%d\n",x, ctxmap, lx,rx);
      debug_flush();
      callout_unknown(rd, w,cols, lx,rx-1, ctxmap);
      goto restart;
    }
  }
  add_result(rd, 0,-1,-1,0);
  fprintf(debug,"OCR  finished %d glyphs\n",rd->nresults);
  debug_flush();
  return rd->results;
}

OcrReader *ocr_init(int h) {
  OcrReader *rd;

  rd= malloc(sizeof(*rd));  eassert(rd);
  memset(rd,0,sizeof(*rd));
  rd->h= h;
  readdb(rd);
  return rd;
}
